/*

 DynFileFmtIO Class for C++
 (c) 2004 Jonathan Bettencourt / Kreative Korporation

 This is the standard code base for manipulating DFF (Dynamic File Format) files.


 This code is under the MIT license.

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in the
 Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef _____DYNFILEFMTIO_____
#define _____DYNFILEFMTIO_____

#include <string>
#include "binarystream.h"
#include "memoryblock.h"
#include "stringmaniplib.h"
using namespace std;

const int DFF_FLAG_BIT_LOCKED = 0;
const int DFF_FLAG_BIT_READ_ONLY = 0;
const int DFF_FLAG_BIT_SYSTEM = 1;
const int DFF_FLAG_BIT_PRELOAD = 2;
const int DFF_FLAG_BIT_PURGABLE = 3;
const int DFF_FLAG_BIT_PURGEABLE = 3;
const int DFF_FLAG_BIT_FROM_FILE = 4;
const int DFF_FLAG_BIT_FROM_RSRC = 5;
const int DFF_FLAG_BIT_FROM_RESOURCE = 5;
const int DFF_FLAG_BIT_HIDDEN = 6;
const int DFF_FLAG_BIT_INVISIBLE = 6;
const int DFF_FLAG_BIT_DISABLED = 7;
const int DFF_FLAG_BIT_PROTECTED = 8;
const int DFF_FLAG_BIT_FIXED = 9;
const int DFF_FLAG_BIT_RESERVED = 10;
const int DFF_FLAG_BIT_COMPRESSED = 11;
const int DFF_FLAG_BIT_APP_USE_1 = 12;
const int DFF_FLAG_BIT_APP_USE_2 = 13;
const int DFF_FLAG_BIT_APP_USE_3 = 14;
const int DFF_FLAG_BIT_APP_USE_4 = 15;

const unsigned long int DFF_INDEX_UNDEFINED = 0xFFFFFFFF;

const unsigned long int UINT32_NEGATIVE_1 = (unsigned long int)(-1);

string dffFlagBitsToUserString(unsigned short int);

class dffrecord
{
	public:

	signed short int datatype;
	bool flagbits[16];
	signed long int idnumber;
	unsigned long int readindatalen;
	string symbol;
	string title;
	memoryblock thedata;

	dffrecord(void);

	unsigned short int getflagint(void);
	void setflagint(unsigned short int);
	bool getlock(void);
	void lock(void);
	void unlock(void);
	string correctedsymbol(void);
};

class dynfilefmtio
{
	public:

	unsigned long int buffersize;
	bool notdff;

	dynfilefmtio(void);
	dynfilefmtio(char *, bool, bool);
	dynfilefmtio(string, bool, bool);
	~dynfilefmtio(void);


	void create(const string &);
	void open(const string &);
	void open(const string &, bool, bool);
	void close(void);


	/* tab stops every 8 spaces */

	int			nthNumber		(const string &, unsigned long int);
	string			nthSymbol		(unsigned long int);
	unsigned long int	numNumbers		(const string &);
	unsigned long int	numSymbols		(void);
	long int		nextAvailableNumber	(const string &);
	long int		nextAvailableNumber	(const string &, long int);
	unsigned long int	maxIndexValue		(void);


	void	addRecord		(dffrecord);
	void	deleteAllRecords	(void);
	void	deleteRecord		(const string &, long int);
	void	deleteRecordByIndex	(unsigned long int);
	void	deleteRecordByName	(const string &, const string &);


	unsigned long int	getRecordIndex		(const string &, long int);
	unsigned long int	getRecordIndexByName	(const string &, const string &);


	dffrecord		getRecord		(const string &, long int);
	memoryblock		getRecordData		(const string &, long int);
	int			getRecordDataType	(const string &, long int);
	unsigned long int	getRecordLength		(const string &, long int);
	unsigned short int	getRecordFlags		(const string &, long int);
	string			getRecordName		(const string &, long int);
	long int		getRecordNumber		(const string &, long int);

	dffrecord		getRecordByIndex	(unsigned long int);
	memoryblock		getRecordDataByIndex	(unsigned long int);
	int			getRecordDataTypeByIndex(unsigned long int);
	unsigned long int	getRecordLengthByIndex	(unsigned long int);
	unsigned short int	getRecordFlagsByIndex	(unsigned long int);
	string			getRecordNameByIndex	(unsigned long int);
	long int		getRecordNumberByIndex	(unsigned long int);
	string			getRecordSymbolByIndex	(unsigned long int);

	dffrecord		getRecordByName		(const string &, const string &);
	memoryblock		getRecordDataByName	(const string &, const string &);
	int			getRecordDataTypeByName	(const string &, const string &);
	unsigned long int	getRecordLengthByName	(const string &, const string &);
	unsigned short int	getRecordFlagsByName	(const string &, const string &);
	string			getRecordNameByName	(const string &, const string &);
	long int		getRecordNumberByName	(const string &, const string &);


	void	setRecord		(const string &, long int,	 dffrecord &);
	void	setRecordData		(const string &, long int,	 memoryblock &);
	void	setRecordDataType	(const string &, long int,	 int);
	void	setRecordFlags		(const string &, long int,	 unsigned short int);
	void	setRecordName		(const string &, long int,	 const string &);
	void	setRecordNumber		(const string &, long int,	 long int);
	void	setRecordSymbol		(const string &, long int,	 const string &);

	void	setRecordByIndex	(unsigned long int,		 dffrecord &);
	void	setRecordDataByIndex	(unsigned long int,		 memoryblock &);
	void	setRecordDataTypeByIndex(unsigned long int,		 int);
	void	setRecordFlagsByIndex	(unsigned long int,		 unsigned short int);
	void	setRecordNameByIndex	(unsigned long int,		 const string &);
	void	setRecordNumberByIndex	(unsigned long int,		 long int);
	void	setRecordSymbolByIndex	(unsigned long int,		 const string &);

	void	setRecordByName		(const string &, const string &, dffrecord &);
	void	setRecordDataByName	(const string &, const string &, memoryblock &);
	void	setRecordDataTypeByName	(const string &, const string &, int);
	void	setRecordFlagsByName	(const string &, const string &, unsigned short int);
	void	setRecordNameByName	(const string &, const string &, const string &);
	void	setRecordNumberByName	(const string &, const string &, long int);
	void	setRecordSymbolByName	(const string &, const string &, const string &);


	private:

	unsigned long int datastartptr;
	unsigned long int headerstartptr;
	unsigned long int namespacestartptr;
	unsigned long int numrecords;
	bool expandorcontractbailed;
	binarystream mystream;

	void init(bool);

	int diff(int, int);

	void contract(unsigned long int, unsigned long int);
	void expand(unsigned long int, unsigned long int);
	unsigned long int recordDataStart(unsigned long int);
	unsigned long int recordHeaderStart(unsigned long int);
	unsigned long int recordNameStart(unsigned long int);
};

inline int dynfilefmtio::diff(int x, int y)
{
	return (x>y)?(x-y):(y-x);
}

#endif